theory Elgamal imports
  Generative_Probabilistic_Value
  Cyclic_Group_SPMF
  Computational_Model
  Diffie_Hellman
  IND_CPA_PK_Single
begin

lemma case_prod_distrib_bind_spmf:
  "bind_spmf p (\<lambda>x. case_prod (f x) y) = case_prod (\<lambda>a b. bind_spmf p (\<lambda>x. f x a b)) y"
by(cases y) simp

subsection {* Elgamal encryption scheme *}

locale elgamal_base =
  fixes \<G> :: "'grp cyclic_group" (structure)
begin

type_synonym 'grp' pub_key = "'grp'"
type_synonym 'grp' priv_key = nat
type_synonym 'grp' plain = 'grp'
type_synonym 'grp' cipher = "'grp' \<times> 'grp'"

definition key_gen :: "('grp pub_key \<times> 'grp priv_key) spmf"
where 
  "key_gen = do {
     x \<leftarrow> sample_uniform (order \<G>);
     return_spmf (\<^bold>g (^) x, x)
  }"

definition aencrypt :: "'grp pub_key \<Rightarrow> 'grp \<Rightarrow> 'grp cipher spmf"
where
  "aencrypt \<alpha> msg = do {
    y \<leftarrow> sample_uniform (order \<G>);
    return_spmf (\<^bold>g (^) y, (\<alpha> (^) y) \<otimes> msg)
  }"

definition adecrypt :: "'grp priv_key \<Rightarrow> 'grp cipher \<Rightarrow> 'grp option"
where
  "adecrypt x = (\<lambda>(\<beta>, \<zeta>). Some (\<zeta> \<otimes> (inv (\<beta> (^) x))))"

abbreviation valid_plain :: "'grp \<Rightarrow> bool"
where "valid_plain msg \<equiv> msg \<in> carrier \<G>"

sublocale ind_cpa: ind_cpa key_gen aencrypt adecrypt valid_plain .
sublocale ddh: ddh \<G> .

fun elgamal_adversary
   :: "('grp pub_key, 'grp plain, 'grp cipher, 'state) ind_cpa.adversary
   \<Rightarrow> 'grp ddh.adversary"
where
  "elgamal_adversary (\<A>1, \<A>2) \<alpha> \<beta> \<gamma> = TRY do {
    b \<leftarrow> coin_spmf;
    ((msg1, msg2), \<sigma>) \<leftarrow> \<A>1 \<alpha>;
    (* have to check that the attacker actually sends an element from the group; otherwise flip a coin *)
    _ :: unit \<leftarrow> assert_spmf (valid_plain msg1 \<and> valid_plain msg2);
    guess \<leftarrow> \<A>2 (\<beta>, \<gamma> \<otimes> (if b then msg1 else msg2)) \<sigma>;
    return_spmf (guess = b)
  } ELSE coin_spmf"

lemma lossless_key_gen [simp]: "lossless_spmf (key_gen) \<longleftrightarrow> 0 < order \<G>"
by(simp add: key_gen_def Let_def)

lemma lossless_aencrypt [simp]:
  "lossless_spmf (aencrypt key plain) \<longleftrightarrow> 0 < order \<G>"
by(simp add: aencrypt_def Let_def)

lemma lossless_elgamal_adversary:
  "\<lbrakk> ind_cpa.lossless \<A>; 0 < order \<G> \<rbrakk>
  \<Longrightarrow> ddh.lossless (elgamal_adversary \<A>)"
by(cases \<A>)(simp add: ddh.lossless_def ind_cpa.lossless_def Let_def split_def)

end

locale elgamal = elgamal_base + cyclic_group \<G> +
  assumes finite_group: "finite (carrier \<G>)"
begin 

lemma advantage_elgamal: "ddh.advantage (elgamal_adversary \<A>) = ind_cpa.advantage \<A>"
proof -
  obtain \<A>1 \<A>2 where [simp]: "\<A> = (\<A>1, \<A>2)" by(cases \<A>)
  note [simp] = order_gt_0_iff_finite finite_group
  have "ddh.ddh_1 (elgamal_adversary \<A>) = TRY do {
       let q = order \<G>; 
       x \<leftarrow> sample_uniform q;
       y \<leftarrow> sample_uniform q;
       ((msg1, msg2), \<sigma>) \<leftarrow> \<A>1 (\<^bold>g (^) x);
       _ :: unit \<leftarrow> assert_spmf (msg1 \<in> carrier \<G> \<and> msg2 \<in> carrier \<G>);
       b \<leftarrow> coin_spmf;
       z \<leftarrow> map_spmf (\<lambda>z. \<^bold>g (^) z \<otimes> (if b then msg1 else msg2)) (sample_uniform q);
       guess \<leftarrow> \<A>2 (\<^bold>g (^) y, z) \<sigma>;
       return_spmf (guess \<longleftrightarrow> b)
     } ELSE coin_spmf"
    apply(clarsimp simp add: Let_def try_spmf_bind_out split_def bind_map_spmf o_def ddh.ddh_1_def)
    apply(rewrite bind_commute_spmf[where p=coin_spmf])
    apply(rewrite bind_commute_spmf[where q="\<A>1 _"])
    apply(rewrite bind_commute_spmf[where q="coin_spmf"])
    apply(rewrite bind_commute_spmf[where q="assert_spmf _"])
    apply(rewrite bind_commute_spmf[where q="assert_spmf _"])
    by(simp)
  also have "\<dots> = TRY do {
       let q = order \<G>;
       x \<leftarrow> sample_uniform q;
       y \<leftarrow> sample_uniform q;
       ((msg1, msg2), \<sigma>) \<leftarrow> \<A>1 (\<^bold>g (^) x);
       _ :: unit \<leftarrow> assert_spmf (msg1 \<in> carrier \<G> \<and> msg2 \<in> carrier \<G>);
       z \<leftarrow> map_spmf (\<lambda>z. \<^bold>g (^) z) (sample_uniform q);
       guess \<leftarrow> \<A>2 (\<^bold>g (^) y, z) \<sigma>;
       map_spmf (op = guess) coin_spmf
     } ELSE coin_spmf"
    by(clarsimp simp add: Let_def sample_uniform_one_time_pad cong: bind_spmf_cong_simp)
      (simp add: map_spmf_conv_bind_spmf bind_commute_spmf[where p=coin_spmf])
  also have "\<dots> = coin_spmf"
    by(simp add: map_eq_const_coin_spmf Let_def split_def bind_spmf_const try_bind_spmf_lossless2' scale_bind_spmf weight_spmf_le_1 scale_scale_spmf del: try_spmf_lossless)
      (simp add: bind_spmf_const)
  also have "ddh.ddh_0 (elgamal_adversary \<A>) = ind_cpa.ind_cpa \<A>"
    unfolding ddh.ddh_0_def IND_CPA_PK_Single.ind_cpa.ind_cpa_def key_gen_def aencrypt_def
    apply(clarsimp simp add: Let_def try_spmf_bind_out split_def)
    apply(rewrite bind_commute_spmf[where q=coin_spmf])
    apply(rewrite bind_commute_spmf[where q="\<A>1 _"])
    apply(rewrite bind_commute_spmf[where q="assert_spmf _"])
    by(simp add: nat_pow_pow eq_commute)
  ultimately show ?thesis unfolding ddh.advantage_def ind_cpa.advantage_def by(simp add: spmf_of_set)
qed

end

end