theory IND_CPA imports
  Generative_Probabilistic_Value
  Computational_Model
  Negligible
  Environment_Functor
begin

subsection {* The IND-CPA security game *}

locale ind_cpa = 
  fixes key_gen :: "'key spmf" -- {*  probabilistic *}
  and encrypt :: "'key \<Rightarrow> 'plain \<Rightarrow> 'cipher spmf"  -- {*  probabilistic *}
  and decrypt :: "'key \<Rightarrow> 'cipher \<Rightarrow> 'plain option" -- {*  deterministic, but not used *}
  and valid_plain :: "'plain \<Rightarrow> bool" -- {* checks whether a plain text is valid, i.e., has the right format *}
begin

text {*
  We cannot incorporate the predicate @{term valid_plain} in the type @{typ "'plain"} of plaintexts,
  because the single @{typ "'plain"} must contain plaintexts for all values of the security parameter,
  as HOL does not have dependent types.  Consequently, the oracle has to ensure that the received
  plaintexts are valid.
*}

type_synonym ('plain', 'cipher', 'state) adversary = 
  "(('plain' \<times> 'plain') \<times> 'state, 'plain', 'cipher') gpv
   \<times> ('cipher' \<Rightarrow> 'state \<Rightarrow> (bool, 'plain', 'cipher') gpv)"
(*type_synonym ('plain', 'cipher', 'state) adversary = "security \<Rightarrow> ('plain', 'cipher', 'state) adversary'"
*)

definition encrypt_oracle :: "'key \<Rightarrow> unit \<Rightarrow> 'plain \<Rightarrow> ('cipher \<times> unit) spmf"
where
  "encrypt_oracle key \<sigma> plain = do {
     cipher \<leftarrow> encrypt key plain;
     return_spmf (cipher, ())
  }"

definition ind_cpa :: "('plain, 'cipher, 'state) adversary \<Rightarrow> bool spmf"
where
  "ind_cpa \<A> = do {
     let (\<A>1, \<A>2) = \<A>;
     key \<leftarrow> key_gen;
     b \<leftarrow> coin_spmf;
     (guess, _) \<leftarrow> exec_gpv (encrypt_oracle key) (do {
         ((m0, m1), \<sigma>) \<leftarrow> \<A>1;
         if valid_plain m0 \<and> valid_plain m1 then do {
           cipher \<leftarrow> lift_spmf (encrypt key (if b then m0 else m1));
           \<A>2 cipher \<sigma>
         } else lift_spmf coin_spmf
       }) ();
     return_spmf (guess = b)
  }"

definition advantage :: "('plain, 'cipher, 'state) adversary \<Rightarrow> real"
where "advantage \<A> = \<bar>spmf (ind_cpa \<A>) True - 1/2\<bar>"

lemma advantage_nonneg: "advantage \<A> \<ge> 0" by(simp add: advantage_def)
(*
abbreviation secure_for :: "('plain, 'cipher, 'state) adversary \<Rightarrow> bool"
where "secure_for \<A> \<equiv> negligible (advantage \<A>)"
*)
definition ibounded_by :: "('plain, 'cipher, 'state) adversary \<Rightarrow> nat \<Rightarrow> bool"
where 
  "ibounded_by = (\<lambda>(\<A>1, \<A>2) q. 
  (\<exists>q1 q2. interaction_bounded_by \<A>1 q1 \<and> (\<forall>cipher \<sigma>. interaction_bounded_by (\<A>2 cipher \<sigma>) q2) \<and> q1 + q2 \<le> q))"

lemma ibounded_byE [consumes 1, case_names ibounded_by, elim?]:
  assumes "ibounded_by (\<A>1, \<A>2) q"
  obtains q1 q2
  where "q1 + q2 \<le> q"
  and "interaction_bounded_by \<A>1 q1"
  and "\<And>cipher \<sigma>. interaction_bounded_by (\<A>2 cipher \<sigma>) q2"
using assms by(auto simp add: ibounded_by_def)

lemma ibounded_byI [intro?]:
  "\<lbrakk> interaction_bounded_by \<A>1 q1; \<And>cipher \<sigma>. interaction_bounded_by (\<A>2 cipher \<sigma>) q2; q1 + q2 \<le> q \<rbrakk>
  \<Longrightarrow> ibounded_by (\<A>1, \<A>2) q"
by(auto simp add: ibounded_by_def)
(*
abbreviation ibounded_by :: "('plain, 'cipher, 'state) adversary \<Rightarrow> (security \<Rightarrow> nat) \<Rightarrow> bool"
where "ibounded_by \<equiv> rel_envir ibounded_by'"
*)
definition lossless :: "('plain, 'cipher, 'state) adversary \<Rightarrow> bool"
where "lossless = (\<lambda>(\<A>1, \<A>2). lossless_gpv \<I>_top \<A>1 \<and> (\<forall>cipher \<sigma>. lossless_gpv \<I>_top (\<A>2 cipher \<sigma>)))"
(*
abbreviation lossless :: "('plain, 'cipher, 'state) adversary \<Rightarrow> bool"
where "lossless \<equiv> pred_envir lossless'"
*)
end

end