section {* Examples *}

theory IND_CPA_PK imports
  Generative_Probabilistic_Value
  Computational_Model
  Negligible
  Environment_Functor
begin

subsection {* The IND-CPA security game for public-key encryption with oracle access *}

locale ind_cpa_pk = 
  fixes key_gen :: "('pubkey \<times> 'privkey, 'call, 'ret) gpv" -- {*  probabilistic *}
  and aencrypt :: "'pubkey \<Rightarrow> 'plain \<Rightarrow> ('cipher, 'call, 'ret) gpv"  -- {*  probabilistic w/ access to an oracle *}
  and adecrypt :: "'privkey \<Rightarrow> 'cipher \<Rightarrow> ('plain, 'call, 'ret) gpv" -- {*  not used *}
  and valid_plain :: "'plain \<Rightarrow> bool" -- {* checks whether a plain text is valid, i.e., has the right format *}
begin

text {*
  We cannot incorporate the predicate @{term valid_plain} in the type @{typ "'plain"} of plaintexts,
  because the single @{typ "'plain"} must contain plaintexts for all values of the security parameter,
  as HOL does not have dependent types.  Consequently, the oracle has to ensure that the received
  plaintexts are valid.
*}

type_synonym ('pubkey', 'plain', 'cipher', 'call', 'ret', 'state) adversary = 
  "('pubkey' \<Rightarrow> (('plain' \<times> 'plain') \<times> 'state, 'call', 'ret') gpv)
   \<times> ('cipher' \<Rightarrow> 'state \<Rightarrow> (bool, 'call', 'ret') gpv)"

fun ind_cpa :: "('pubkey, 'plain, 'cipher, 'call, 'ret, 'state) adversary \<Rightarrow> (bool, 'call, 'ret) gpv"
where
  "ind_cpa (\<A>1, \<A>2) = TRY do {
     (pk, sk) \<leftarrow> key_gen;
     b \<leftarrow> lift_spmf coin_spmf;
     ((m0, m1), \<sigma>) \<leftarrow> (\<A>1 pk);
     assert_gpv (valid_plain m0 \<and> valid_plain m1);
     cipher \<leftarrow> aencrypt pk (if b then m0 else m1);
     guess \<leftarrow> \<A>2 cipher \<sigma>; 
     Done (guess = b)
   } ELSE lift_spmf coin_spmf"

definition advantage :: "('\<sigma> \<Rightarrow> 'call \<Rightarrow> ('ret \<times> '\<sigma>) spmf) \<Rightarrow> '\<sigma> \<Rightarrow> ('pubkey, 'plain, 'cipher, 'call, 'ret, 'state) adversary \<Rightarrow> real"
where "advantage oracle \<sigma> \<A> = \<bar>spmf (run_gpv oracle (ind_cpa \<A>) \<sigma>) True - 1/2\<bar>"

lemma advantage_nonneg: "advantage oracle \<sigma> \<A> \<ge> 0" by(simp add: advantage_def)

definition ibounded_by :: "('pubkey, 'plain, 'cipher, 'call, 'ret, 'state) adversary \<Rightarrow> nat \<Rightarrow> bool"
where 
  "ibounded_by = (\<lambda>(\<A>1, \<A>2) q. 
  (\<exists>q1 q2. (\<forall>pk. interaction_bounded_by (\<A>1 pk) q1) \<and> (\<forall>cipher \<sigma>. interaction_bounded_by (\<A>2 cipher \<sigma>) q2) \<and> q1 + q2 \<le> q))"

lemma ibounded_by'E [consumes 1, case_names ibounded_by', elim?]:
  assumes "ibounded_by (\<A>1, \<A>2) q"
  obtains q1 q2
  where "q1 + q2 \<le> q"
  and "\<And>pk. interaction_bounded_by (\<A>1 pk) q1"
  and "\<And>cipher \<sigma>. interaction_bounded_by (\<A>2 cipher \<sigma>) q2"
using assms by(auto simp add: ibounded_by_def)

lemma ibounded_byI [intro?]:
  "\<lbrakk> \<And>pk. interaction_bounded_by (\<A>1 pk) q1; \<And>cipher \<sigma>. interaction_bounded_by (\<A>2 cipher \<sigma>) q2; q1 + q2 \<le> q \<rbrakk>
  \<Longrightarrow> ibounded_by (\<A>1, \<A>2) q"
by(auto simp add: ibounded_by_def)

definition lossless :: "('pubkey, 'plain, 'cipher, 'call, 'ret, 'state) adversary \<Rightarrow> bool"
where "lossless = (\<lambda>(\<A>1, \<A>2). (\<forall>pk. lossless_gpv \<I>_top (\<A>1 pk)) \<and> (\<forall>cipher \<sigma>. lossless_gpv \<I>_top (\<A>2 cipher \<sigma>)))"

end

end